using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using CsOde;

namespace XnaRagdoll
{
    #region typedef
    using WorldID = IntPtr;
    using SpaceID = IntPtr;
    using BodyID = IntPtr;
    using GeomID = IntPtr;
    using JointID = IntPtr;
    using JointGroupID = IntPtr;
    using HeightfieldDataID = IntPtr;
    using TriMeshDataID = IntPtr;
    using Real = Single;
    #endregion

    /// <summary>
    /// ֐ߑ̃m[hNX
    /// </summary>
    public class ArticulatedBodyNode : IPhysicalObject, IDisposable
    {
        #region field
        /// <summary>em[h</summary>
        private ArticulatedBodyNode parent = null;
        /// <summary>֐߁iem[h֐ڑj</summary>
        private JointID joint = JointID.Zero;
        /// <summary>֐߂̎</summary>
        /// <remarks>dJointTypeFixedCdJointTypeHingeCdJointTypeUniversalCdJointTypeBall ̂݃T|[g</remarks>
        private Ode.JointType jointType = Ode.JointType.None;
        /// <summary>{[</summary>
        private PhysicalPrimitive bone = null;
        /// <summary>qm[h</summary>
        private List<ArticulatedBodyNode> children = new List<ArticulatedBodyNode>();
        #endregion

        #region property
        /// <summary>em[h</summary>
        public ArticulatedBodyNode Parent
        {
            get { return parent; }
            protected set { parent = value; }
        }
        /// <summary>֐߁iem[h֐ڑj</summary>
        public JointID Joint
        {
            get { return joint; }
            protected set { joint = value; }
        }
        /// <summary>֐߂̎</summary>
        public Ode.JointType JointType
        {
            get { return jointType; }
            protected set { jointType = value; }
        }
        /// <summary>{[</summary>
        public PhysicalPrimitive Bone
        {
            get { return bone; }
        }
        /// <summary>qm[h̐</summary>
        public int NumChildren
        {
            get { return children.Count; }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="bone">em[h</param>
        public ArticulatedBodyNode(PhysicalPrimitive bone)
        {
            this.bone = bone;
        }

        /// <summary>
        /// Dispose\bh
        /// </summary>
        public virtual void Dispose()
        {
            for (int i = 0; i < NumChildren; ++i)
                children[i].Dispose();
            children.Clear();

            if (joint != JointID.Zero)
                Ode.JointDestroy(joint);
            joint = JointID.Zero;

            if (bone != null)
                bone.Dispose();
            bone = null;

            jointType = Ode.JointType.None;
            parent = null;
        }

        /// <summary>
        /// qm[h擾
        /// </summary>
        /// <param name="index">CfNX</param>
        /// <returns>qm[h</returns>
        public ArticulatedBodyNode GetChild(int index)
        {
            return children[index];
        }

        /// <summary>
        /// qN̒ǉ
        /// </summary>
        /// <param name="child">qN</param>
        protected void AddChild(ArticulatedBodyNode child)
        {
            children.Add(child);
            child.parent = this;
        }

        /// <summary>
        /// Œ֐߂ŐeNɐڑ
        /// </summary>
        /// <param name="parent">eN</param>
        public virtual void ConnectByFix(ArticulatedBodyNode parent)
        {
            if (joint != JointID.Zero || this.parent != null)
                throw new ApplicationException();

            joint = Ode.JointCreateFixed(Physics.World, JointGroupID.Zero);
            Ode.JointAttach(joint, parent.Bone.Body, bone.Body);
            Ode.JointSetFixed(joint);

            jointType = Ode.JointType.Fixed;

            parent.AddChild(this);
        }

        /// <summary>
        /// ʊ֐߂ŐeNɐڑ
        /// </summary>
        /// <param name="parent">eN</param>
        /// <param name="anchor">ڑ_W</param>
        public virtual void ConnectByBall(ArticulatedBodyNode parent, Vector3 anchor)
        {
            if (joint != JointID.Zero || this.parent != null)
                throw new ApplicationException();

            joint = Ode.JointCreateBall(Physics.World, JointGroupID.Zero);
            Ode.JointAttach(joint, parent.Bone.Body, bone.Body);
            Ode.JointSetBallAnchor(joint, anchor.X, anchor.Y, anchor.Z);

            jointType = Ode.JointType.Ball;

            parent.AddChild(this);
        }

        /// <summary>
        /// qW֐߂ŐeNɐڑ
        /// </summary>
        /// <param name="parent">eN</param>
        /// <param name="anchor">ڑ_W</param>
        /// <param name="axis">]PʃxNg</param>
        /// <param name="limit">֐߉i[0]: C[1]: Cradianj</param>
        public virtual void ConnectByHinge(ArticulatedBodyNode parent, Vector3 anchor, Vector3 axis, float[] limit)
        {
            if (joint != JointID.Zero || this.parent != null)
                throw new ApplicationException();

            joint = Ode.JointCreateHinge(Physics.World, JointGroupID.Zero);
            Ode.JointAttach(joint, parent.Bone.Body, bone.Body);
            Ode.JointSetHingeAnchor(joint, anchor.X, anchor.Y, anchor.Z);
            Ode.JointSetHingeAxis(joint, axis.X, axis.Y, axis.Z);

            Ode.JointSetHingeParam(joint, (int)Ode.JointParam.LoStop, limit[0]);
            Ode.JointSetHingeParam(joint, (int)Ode.JointParam.HiStop, limit[1]);

            jointType = Ode.JointType.Hinge;

            parent.AddChild(this);
        }

        /// <summary>
        /// jo[T֐߂ŐeNɐڑ
        /// </summary>
        /// <param name="parent">eN</param>
        /// <param name="anchor">ڑ_W</param>
        /// <param name="axis1">֐1̉]PʃxNg</param>
        /// <param name="limit1">֐1̉i[0]: C[1]: Cradianj</param>
        /// <param name="axis2">֐2̉]PʃxNg</param>
        /// <param name="limit2">֐2̉i[0]: C[1]: Cradianj</param>
        public virtual void ConnectByUniversal(ArticulatedBodyNode parent, Vector3 anchor, Vector3 axis1, float[] limit1, Vector3 axis2, float[] limit2)
        {
            if (joint != JointID.Zero || this.parent != null)
                throw new ApplicationException();
            if (limit1.Length < 2 || limit2.Length < 2)
                throw new ApplicationException();

            joint = Ode.JointCreateUniversal(Physics.World, JointGroupID.Zero);
            Ode.JointAttach(joint, parent.Bone.Body, bone.Body);
            Ode.JointSetUniversalAnchor(joint, anchor.X, anchor.Y, anchor.Z);

            Ode.JointSetUniversalAxis1(joint, axis1.X, axis1.Y, axis1.Z);
            Ode.JointSetUniversalAxis2(joint, axis2.X, axis2.Y, axis2.Z);

            Ode.JointSetUniversalParam(joint, (int)Ode.JointParam.LoStop, limit1[0]);
            Ode.JointSetUniversalParam(joint, (int)Ode.JointParam.HiStop, limit1[1]);
            Ode.JointSetUniversalParam(joint, (int)Ode.JointParam.LoStop2, limit2[0]);
            Ode.JointSetUniversalParam(joint, (int)Ode.JointParam.HiStop2, limit2[1]);

            jointType = Ode.JointType.Universal;

            parent.AddChild(this);
        }
    }
}
