using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace XnaRagdoll
{
    public class BasicGeometry
    {
        public static Quaternion QuatLn(Quaternion q)
        {
            double theta = Math.Acos(q.W);
            double invSinTheta = theta / Math.Sin(theta);

            Quaternion qe = q * (float)invSinTheta;
            qe.W = 0.0f;
            return qe;
        }
        public static void QuatLn(ref Quaternion q, out Quaternion lnq)
        {
            double theta = Math.Acos(q.W);
            double invSinTheta = theta / Math.Sin(theta);
            lnq = q * (float)invSinTheta;
            lnq.W = 0.0f;
        }
        public static Quaternion QuatExp(Quaternion lnq)
        {
            double theta = Math.Sqrt(lnq.X * lnq.X + lnq.Y * lnq.Y + lnq.Z * lnq.Z);
            double sinTheta = Math.Sin(theta) / theta;
            Quaternion q = lnq * (float)sinTheta;
            q.W = (float)Math.Cos(theta);
            return q;
        }
        public static void QuatExp(ref Quaternion lnq, out Quaternion q)
        {
            double theta = Math.Sqrt(lnq.X * lnq.X + lnq.Y * lnq.Y + lnq.Z * lnq.Z);
            double sinTheta = Math.Sin(theta) / theta;
            q = lnq * (float)sinTheta;
            q.W = (float)Math.Cos(theta);
        }

        public static Matrix Orthogonalize(Matrix src)
        {
            Vector3 v1 = new Vector3(src.M11, src.M12, src.M13);
            Vector3 v2 = new Vector3(src.M21, src.M22, src.M23);
            Vector3 v3 = new Vector3(src.M31, src.M32, src.M33);

            v1.Normalize();
            v2 = v2 - v1 * Vector3.Dot(v2, v1);
            v2.Normalize();
            v3 = v3 - v1 * Vector3.Dot(v1, v3) - v2 * Vector3.Dot(v2, v3);
            v3.Normalize();

            Matrix dest = Matrix.Identity;
            dest.M11 = v1.X; dest.M12 = v1.Y; dest.M13 = v1.Z;
            dest.M21 = v2.X; dest.M22 = v2.Y; dest.M23 = v2.Z;
            dest.M31 = v3.X; dest.M32 = v3.Y; dest.M33 = v3.Z;
            return dest;
        }
    }
}
