using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using CsOde;

namespace XnaRagdoll
{
    #region typedef
    using WorldID = IntPtr;
    using SpaceID = IntPtr;
    using BodyID = IntPtr;
    using GeomID = IntPtr;
    using JointID = IntPtr;
    using JointGroupID = IntPtr;
    using HeightfieldDataID = IntPtr;
    using TriMeshDataID = IntPtr;
    using Real = Single;
    #endregion

    /// <summary>
    /// SĂ̕IuWFNg̊{NX
    /// </summary>
    abstract public class PhysicalPrimitive : IPhysicalObject, IDisposable
    {
        #region field
        /// <summary>ODE - {fB</summary>
        protected BodyID body = BodyID.Zero;
        /// <summary>ODE - WIg</summary>
        protected GeomID geom = GeomID.Zero;
        /// <summary>XNA - \f</summary>
        protected Model model = null;
        /// <summary>PʃTCY{v~eBu̕\XP[</summary>
        protected Vector3 scale = Vector3.Zero;
        /// <summary> [kg]</summary>
        protected float weight = 0.0f;
        #endregion

        #region property
        /// <summary>IuWFNg̈ʒui[hWnj</summary>
        public Vector3 Position
        {
            get
            {
                Ode.Vector3 v;
                Ode.BodyCopyPosition(body, out v);
                return new Vector3(v.X, v.Y, v.Z);
            }
            set
            {
                Ode.BodySetPosition(body, value.X, value.Y, value.Z);
            }
        }
        /// <summary>IuWFNg̕i[hWnj</summary>
        public Matrix Orientation
        {
            get
            {
                Ode.Matrix3 r;
                Ode.BodyCopyRotation(body, out r);

                Matrix m = Matrix.Identity;
                m.M11 = r.M11;
                m.M12 = r.M12;
                m.M13 = r.M13;
                m.M21 = r.M21;
                m.M22 = r.M22;
                m.M23 = r.M23;
                m.M31 = r.M31;
                m.M32 = r.M32;
                m.M33 = r.M33;

                return m;
            }
            set
            {
                Ode.Matrix3 r;
                r.M11 = value.M11;
                r.M12 = value.M12;
                r.M13 = value.M13;
                r.M41 = 0;
                r.M21 = value.M21;
                r.M22 = value.M22;
                r.M23 = value.M23;
                r.M42 = 0;
                r.M31 = value.M31;
                r.M32 = value.M32;
                r.M33 = value.M33;
                r.M43 = 0;

                Ode.BodySetRotation(body, ref r);
            }
        }
        /// <summary>IuWFNg̃gXtH[i[hWnj</summary>
        public Matrix Coordinate
        {
            get
            {
                Vector3 pos = Position;
                Matrix rot = Orientation;
                rot.M41 = pos.X;
                rot.M42 = pos.Y;
                rot.M43 = pos.Z;
                return rot;
            }
            set
            {
                Vector3 pos = new Vector3(value.M41, value.M42, value.M43);
                Matrix rot = BasicGeometry.Orthogonalize(value);
                Position = pos;
                Orientation = rot;
            }
        }
        /// <summary>ODE - {fB</summary>
        public BodyID Body
        {
            get { return body; }
        }
        /// <summary>ODE - WIg</summary>
        public GeomID Geom
        {
            get { return geom; }
        }
        /// <summary>XNA - \f</summary>
        public Model Model
        {
            get { return model; }
        }
        /// <summary> [kg]</summary>
        public float Weight
        {
            get { return weight; }
        }
        /// <summary>PʃTCY{v~eBu̕\XP[</summary>
        public Vector3 Scale
        {
            get { return scale; }
        }
        /// <summary>XP[gXtH[</summary>
        public Matrix ScaleTransform
        {
            get { return Matrix.CreateScale(scale); }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="weight"> [kg]</param>
        /// <param name="scale">\XP[</param>
        /// <param name="model">\f</param>
        public PhysicalPrimitive(float weight, Vector3 scale, Model model)
        {
            this.weight = weight;
            this.scale = scale;
            this.model = model;
        }

        /// <summary>
        /// Dispose\bh
        /// </summary>
        public virtual void Dispose()
        {
            if (body != BodyID.Zero)
                Ode.BodyDestroy(body);
            body = BodyID.Zero;

            if (geom != GeomID.Zero)
                Ode.GeomDestroy(geom);
            geom = GeomID.Zero;
        }
    }

    /// <summary>
    /// `IuWFNg
    /// </summary>
    class PhysicalSphere : PhysicalPrimitive
    {
        #region field
        /// <summary>̔a [m]</summary>
        private readonly float radius = 0;
        #endregion

        #region property
        /// <summary>̔a [m]</summary>
        public float Radius
        {
            get { return radius; }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="weight"> [kg]</param>
        /// <param name="model">\f</param>
        /// <param name="radius">a [m]</param>
        public PhysicalSphere(float weight, Model model, float radius)
            : base(weight, new Vector3(radius, radius, radius), model)
        {
            this.radius = radius;

            Ode.Mass mass;
            Ode.MassSetZero(out mass);
            Ode.MassSetSphereTotal(out mass, weight, Radius);

            body = Ode.BodyCreate(Physics.World);
            Ode.BodySetMass(body, ref mass);

            geom = Ode.CreateSphere(Physics.Space, radius);
            Ode.GeomSetBody(geom, body);
        }
    }

    /// <summary>
    /// ̕IuWFNg
    /// </summary>
    class PhysicalBox : PhysicalPrimitive
    {
        #region field
        /// <summary>̂̊eӂ̒</summary>
        private readonly Vector3 size = Vector3.Zero;
        #endregion

        #region property
        /// <summary>̊eӂ̒</summary>
        public Vector3 Size
        {
            get { return size; }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="weight"> [kg]</param>
        /// <param name="model">\f</param>
        /// <param name="size">̃TCY [m]</param>
        public PhysicalBox(float weight, Model model, Vector3 size)
            : base(weight, size, model)
        {
            this.size = size;

            Ode.Mass mass;
            Ode.MassSetZero(out mass);
            Ode.MassSetBoxTotal(out mass, weight, size.X, size.Y, size.Z);

            body = Ode.BodyCreate(Physics.World);
            Ode.BodySetMass(body, ref mass);

            geom = Ode.CreateBox(Physics.Space, size.X, size.Y, size.Z);
            Ode.GeomSetBody(geom, body);
        }
    }

    /// <summary>
    /// ~IuWFNg
    /// </summary>
    class PhysicalCylinder : PhysicalPrimitive
    {
        #region field
        /// <summary> [m]</summary>
        private readonly float length;
        /// <summary>a [m]</summary>
        private readonly float radius;
        #endregion

        #region property
        /// <summary> [m]</summary>
        public float Length
        {
            get { return length; }
        }
        /// <summary>a [m]</summary>
        public float Radius
        {
            get { return radius; }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="weight"> [kg]</param>
        /// <param name="model">\f</param>
        /// <param name="length">~</param>
        /// <param name="radius">~a</param>
        public PhysicalCylinder(float weight, Model model, float length, float radius)
            : base(weight, new Vector3(radius, radius, length), model)
        {
            this.length = length;
            this.radius = radius;

            Ode.Mass mass;
            Ode.MassSetCylinderTotal(out mass, weight, 3, radius, length);

            body = Ode.BodyCreate(Physics.World);
            Ode.BodySetMass(body, ref mass);

            geom = Ode.CreateCylinder(Physics.Space, radius, length);
            Ode.GeomSetBody(geom, body);
        }
    }

    /// <summary>
    /// nʕIuWFNg
    /// </summary>
    /// <remarks>
    /// ڐGɂ̂ݗp̂ŃWIĝ݂tB[hɎD
    /// ̂߁CPhysicalObjectNXpȂD
    /// </remarks>
    class PhysicalGround : IDisposable
    {
        #region field
        /// <summary>ʃp[^ (a, b, c, d)Cax + by + cz + d = 0</summary>
        private readonly Vector4 planeParam = Vector4.Zero;
        /// <summary>ODE - nʃWIg</summary>
        protected GeomID geom = GeomID.Zero;
        #endregion

        #region property
        /// <summary>ʃp[^</summary>
        public Vector4 PlaneParam
        {
            get { return planeParam; }
        }
        /// <summary>ODE - WIg</summary>
        public GeomID Geom
        {
            get { return geom; }
        }
        #endregion

        /// <summary>
        /// RXgN^
        /// </summary>
        /// <param name="planeParam">ʃp[^</param>
        public PhysicalGround(Vector4 planeParam)
        {
            this.planeParam = planeParam;
            geom = Ode.CreatePlane(Physics.Space, planeParam.X, planeParam.Y, planeParam.Z, planeParam.W);
        }

        /// <summary>
        /// Dispose\bh
        /// </summary>
        public virtual void Dispose()
        {
            if (geom != GeomID.Zero)
                Ode.GeomDestroy(geom);
            geom = GeomID.Zero;
        }
    }
}
