using System;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using CsOde;

namespace XnaRagdoll
{
    #region typedef
    using WorldID = IntPtr;
    using SpaceID = IntPtr;
    using BodyID = IntPtr;
    using GeomID = IntPtr;
    using JointID = IntPtr;
    using JointGroupID = IntPtr;
    using HeightfieldDataID = IntPtr;
    using TriMeshDataID = IntPtr;
    using Real = Single;
    #endregion

    /// <summary>
    /// IuWFNgC^[tFCX
    /// </summary>
    public interface IPhysicalObject
    {
    }

    /// <summary>
    /// V~[VԃNXiÓINXj
    /// </summary>
    public static class Physics
    {
        #region delegate
        public delegate void StepEvent();
        #endregion

        #region field
        /// <summary>ODE - [hID</summary>
        private static WorldID world = WorldID.Zero;
        /// <summary>ODE - Xy[XID</summary>
        private static SpaceID space = SpaceID.Zero;
        /// <summary>ODE - ڐGWCgO[vID</summary>
        private static JointGroupID contactGroup = JointGroupID.Zero;
        /// <summary>Փ˔R[obN</summary>
        private static Ode.NearCallback defaultCollisionCallback = null;
        #endregion

        #region property
        /// <summary>ODE - [hID</summary>
        public static WorldID World
        {
            get { return world; }
            private set { space = value; }
        }
        /// <summary>ODE - Xy[XID</summary>
        public static SpaceID Space
        {
            get { return space; }
            private set { space = value; }
        }
        /// <summary>ODE - ڐGWCgO[vID</summary>
        public static JointGroupID ContactGroup
        {
            get { return contactGroup; }
            private set { contactGroup = value; }
        }
        /// <summary>d͉xxNg</summary>
        public static Vector3 Gravity
        {
            get
            {
                Ode.Vector3 g;
                Ode.WorldGetGravity(world, out g);
                return new Vector3(g.X, g.Y, g.Z);
            }
        }
        /// <summary>Փ˔R[obN</summary>
        public static Ode.NearCallback DefaultCollisionCallback
        {
            get { return defaultCollisionCallback; }
        }
        #endregion

        /// <summary>
        /// ODEԂ̍\z
        /// </summary>
        public static void Create(Vector3 gravity)
        {
            if (world != WorldID.Zero || space != SpaceID.Zero)
                return;

            Ode.InitODE();

            world = Ode.WorldCreate();
            Ode.WorldSetGravity(world, gravity.X, gravity.Y, gravity.Z);

            space = Ode.HashSpaceCreate(SpaceID.Zero);
            contactGroup = Ode.JointGroupCreate(0);

            defaultCollisionCallback = new Ode.NearCallback(CollisionCallback);
        }

        /// <summary>
        /// ODEԂ̔j
        /// </summary>
        public static void Destroy()
        {
            Ode.JointGroupDestroy(contactGroup);
            Ode.SpaceDestroy(space);
            Ode.WorldDestroy(world);

            contactGroup = JointGroupID.Zero;
            space = SpaceID.Zero;
            contactGroup = JointGroupID.Zero;

            Ode.CloseODE();
        }

        /// <summary>
        /// ODEԂ̎gp̉ۃ`FbN
        /// </summary>
        /// <returns>true: V~[Vs\Cfalse: Ԗ\z</returns>
        public static bool IsActive()
        {
            return (world != WorldID.Zero) && (space != SpaceID.Zero);
        }

        /// <summary>
        /// V~[V1Xebvi߂
        /// </summary>
        /// <param name="spf">V~[Vݕ [sec/frame]</param>
        /// <param name="proc">Փ˔p̃R[obNfQ[g</param>
        public static void Step(Real spf, Ode.NearCallback proc)
        {
            Ode.SpaceCollide(space, IntPtr.Zero, proc);
            Ode.WorldStep(world, spf);
            Ode.JointGroupEmpty(contactGroup);
        }

        /// <summary>
        /// ftHg̏Փ˔R[obNfQ[g
        /// </summary>
        /// <param name="data"></param>
        /// <param name="geom0">WIg0</param>
        /// <param name="geom1">WIg1</param>
        private static void CollisionCallback(IntPtr data, GeomID geom0, GeomID geom1)
        {
            const int N = 20;
            Ode.ContactGeom[] contacts = new Ode.ContactGeom[N];

            BodyID b0 = Ode.GeomGetBody(geom0);
            BodyID b1 = Ode.GeomGetBody(geom1);

            if (b0 != BodyID.Zero && b1 != BodyID.Zero && Ode.AreConnected(b0, b1) == true)
                return;
            if (b0 != BodyID.Zero && b1 != BodyID.Zero && Ode.AreConnectedExcluding(b0, b1, Ode.JointType.Contact) == true)
                return;

            Ode.Contact contact = new Ode.Contact();
            contact.surface.mode = Ode.ContactFlags.Bounce;
            contact.surface.mu = Real.MaxValue;
            contact.surface.mu2 = Real.MaxValue;
            contact.surface.bounce = 0.5f;
            contact.surface.bounce_vel = 0.1f;

            int n = Ode.Collide(geom0, geom1, N, contacts, Ode.ContactGeom.SizeOf);
            for (int i = 0; i < n; ++i)
            {
                contact.geom = contacts[i];
                JointID c = Ode.JointCreateContact(world, contactGroup, ref contact);
                Ode.JointAttach(c, b0, b1);
            }
        }
    }
}
